from typing import Callable, Tuple

def copy(s: str) -> None: ...
def paste() -> str: ...
def determine_clipboard() -> Tuple[Callable[[str], None], Callable[[], str]]: ...

class PyperclipException(Exception): ...
